type op = Add | Sub | Mult | Div | Equal | Neq | Less | Leq | Greater | Geq (*| And_ | Or_*)

type prim = Int | Float | String

type signature =
	Signature of (string * prim) array

type mandatorylit = 
	  LitInt of int
	| LitFloat of float
	| LitString of string

type tablerow =
	LiteralRow of mandatorylit array

type tabledecl = 
		FullTableLiteral of string * signature * tablerow array
  | SingleVarLiteral of (string * prim) * mandatorylit

type expr =
    Literal of mandatorylit
  | Id of string
  | Binop of expr * op * expr
  | Assign of string * expr
  | TAssign of string * string * expr
  | Call of string * expr list
  | GetCol of string * string
(*  | Filter of string * expr *)
  | Noexpr

type stmt =
    Block of stmt list
  | Expr of expr
  | Return of expr
  | If of expr * stmt * stmt
  | For of expr * expr * expr * stmt
  | While of expr * stmt
  | Pipe of string * string * stmt list
  | LocalDecl of tabledecl
  | RowAppend of string * string
  | ColAppend of string * string


type func_decl = {
    fname : string;
    returnsig : tabledecl;
    formals : (string * prim) list;
    body : stmt list;
  }

type program = tabledecl list * func_decl list

let string_of_vdecl id = "int " ^ id ^ ";\n"

let string_of_mandatory_lit = function
	  LitInt(i) -> string_of_int(i)
	| LitFloat(f) -> string_of_float(f)
	| LitString(s) -> if s = "" then "\"\"" else s
		
let string_of_tablerow = function
		LiteralRow(lits) -> String.concat ", " (Array.to_list (Array.map string_of_mandatory_lit lits))

let string_of_tabletype = function
	(s, p) -> ((match p with Int -> "int" | Float -> "float" | String -> "string") ^ " " ^ s)

let string_of_signature = function
		Signature(sigs) -> String.concat ", " (Array.to_list (Array.map string_of_tabletype sigs))

let string_of_tdecl = function
  	FullTableLiteral(s1, sig_n, tablerows) -> "table " ^ s1 ^ " = #{\n" ^ (string_of_signature sig_n) ^ "\n#" ^ (String.concat "\n#" (Array.to_list (Array.map string_of_tablerow tablerows))) ^ "}#\n"
  | SingleVarLiteral((sname, sprim), lv) -> (match sprim with Int -> "int" | Float -> "float" 
		| String -> "string") ^ " " ^ sname ^ " = " ^ string_of_mandatory_lit(lv) ^ ";\n"

let rec string_of_expr = function
    Literal(l) -> string_of_mandatory_lit l
  | Id(s) -> s
  | Binop(e1, o, e2) ->
      string_of_expr e1 ^ " " ^
      (match o with
	Add -> "+" | Sub -> "-" | Mult -> "*" | Div -> "/"
      | Equal -> "==" | Neq -> "!="
      | Less -> "<" | Leq -> "<=" | Greater -> ">" | Geq -> ">=" 
      (*| And_ -> "&&" | Or_ -> "||"*)) ^ " " ^
      string_of_expr e2
  | Assign(v, e) -> v ^ " = " ^ string_of_expr e
  | TAssign(v, v', e) -> v ^ "." ^ v' ^ " = " ^ string_of_expr e
  | Call(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | Noexpr -> ""
  | GetCol(s1, s2) -> s1 ^ "." ^ s2
(*  | Filter(s1, ex2) -> s1 ^ "[" ^ string_of_expr ex2 ^ "]" *)
  
let rec string_of_stmt = function
    Block(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | Expr(expr) -> string_of_expr expr ^ ";\n";
  | Return(expr) -> "return " ^ string_of_expr expr ^ ";\n";
  | If(e, s, Block([])) -> "if (" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s
  | If(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n" ^
      string_of_stmt s1 ^ "else\n" ^ string_of_stmt s2
  | For(e1, e2, e3, s) ->
      "for (" ^ string_of_expr e1  ^ " ; " ^ string_of_expr e2 ^ " ; " ^
      string_of_expr e3  ^ ") " ^ string_of_stmt s
  | While(e, s) -> "while (" ^ string_of_expr e ^ ") " ^ string_of_stmt s
  | Pipe(s1, s2, stmts) -> s1 ^ "|" ^ s2 ^ "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | LocalDecl(td) -> string_of_tdecl td
	| RowAppend(s1,s2) -> s1 ^ "<<" ^ s2
  | ColAppend(s1, s2) -> s1 ^ "::" ^ s2


let string_of_fdecl fdecl =
  "function " ^ (string_of_tdecl fdecl.returnsig) ^ " " ^ fdecl.fname ^ " (" ^ (String.concat ", " (List.map (fun (s, p) -> (match p with Int -> "int" | Float -> "float" | String -> "string") ^ " " ^ s ^ " ") fdecl.formals)) ^ ")\n{\n" ^
  String.concat "" (List.map string_of_stmt fdecl.body) ^
  "}\n"

let string_of_program (vars, funcs) =
	let srav = List.rev vars in
  String.concat "" (List.map string_of_tdecl srav) ^ "\n" ^
  String.concat "\n" (List.map string_of_fdecl funcs)
