open Ast

module StringMap = Map.Make(String)

type env = {
    function_index : Ast.func_decl StringMap.t; (* created once; mapping of function name to function AST; prob could be a list*)
    mutable global_index   : Ast.tabledecl StringMap.t; (* created once but changed; mapping of table name to table AST; necessary *)
    mutable local_index    : (Ast.tabledecl * Ast.func_decl) StringMap.t; (* dynamically updated; mapping of local table name to AST; necessary *)
    mutable current_enclosing_function : string;
  }
																																
exception Illegal_table_param of Ast.tabledecl
exception Function_not_found of string

let prim_check_of_literal mandatoryval = (match mandatoryval with LitInt(_) -> Int | LitString(_) -> String | LitFloat(_) -> Float)

let prim_check_of_table tbsignature littable = 
	let signa = 
		(match tbsignature with Signature(sigNa) -> sigNa) 
		in 
		for j = 0 to ((Array.length littable) - 1) do (
			let litrow = 
				(match littable.(j) with LiteralRow(litr) -> litr) 
			in
			if (Array.length signa <> Array.length litrow ) then
				raise(Failure("Dimension mismatch: " ^ string_of_int j)) 
			else (for i = 0 to ((Array.length signa) - 1) do (
				if ((match signa.(i) with 
						(_, Int) -> Int
					| (_, Float) -> Float
					| (_, String) -> String ) <> 
					(prim_check_of_literal litrow.(i) ) ) 
					then (raise(Failure("Invalid literal in column " ^ string_of_int i ^ "; row " ^ string_of_int j))) 
					else ()
				) done 
			)
		) done

let rec tdecl_list_build map = function
	|	[] -> map
	| SingleVarLiteral((sname, sprim), oplit) as hd :: tl -> let newmap = StringMap.add sname hd map in tdecl_list_build newmap tl
	| FullTableLiteral(tbname, tbsig, tbllit) as hd :: tl -> let newmap = (*Array.fold_left (fun m (s, p) -> StringMap.add s (match p with String_ -> SignatureString(tbname) | Int_ -> SignatureInt(tbname) | Float_ -> SignatureFloat(tbname)) m) map (match tbsig with Signature(sigN) -> sigN)*) StringMap.add tbname hd map in tdecl_list_build newmap tl(*(StringMap.add tbname (TableVar(tbsig)) newmap)*)


let rec fdecl_list_build map = function
	|	[] -> map
	| hd :: tl -> let newmap = StringMap.add hd.fname hd map in fdecl_list_build newmap tl
  

(** Translate a program in AST form into a JavaScript program.  Throw an
    exception if something is wrong, e.g., a reference to an unknown
    variable or function *)
let translate (globals, functions) = 
	(* We need to do two things here.  We need to store the AST tabledecls in a
	data structure so that we can reference it later.  We also need to output of
	the declarations as part of a java class.*)

	let global_table_decls = tdecl_list_build StringMap.empty (List.rev globals) in
	let global_func_decls = fdecl_list_build StringMap.empty (List.rev functions) in

	let rec translate_globals = function
	| [] -> []
	| SingleVarLiteral((nm, prm), lit) :: tl -> 
		if((prim_check_of_literal lit) <> prm) then
			raise(Failure("global: " ^ nm ^ " is assigned a literal that does not match it's declared type"))
		else
			let javaout = "var " ^ nm ^ " = " ^ Ast.string_of_mandatory_lit(lit) ^ ";\n" 
			in javaout :: translate_globals tl
	| FullTableLiteral(nm, tb_sig, tablerows) :: tl -> 
		let () = 
			(try (prim_check_of_table tb_sig tablerows ) with Failure(msg) -> 
				raise(Failure("Error in: " ^ nm ^ ":" ^ msg))) 
		and javaout = "var " ^ nm ^ " = {};\n" ^ 
			nm ^ ".nameString = \"" ^ nm ^ "\";\n" ^
			nm ^ ".rows = new Array();\n" ^
			nm ^ ".headnames = new Array();\n" ^
			nm ^ ".headtypes = new Array();\n" ^
			(* Generate the signature string *)
			nm ^ ".signature = \"" ^ (String.concat "::" (Array.to_list (Array.map (fun (id', prm') -> id' ^ ":" ^
																										(match prm' with Int -> "int" | Float -> "float" | String -> "string")) 
																									(match tb_sig with Signature(sigsig) -> sigsig)
			))) ^ "\";\n" ^
			(* Generate the code to populate the .headnames and .headtypes array *)
			(String.concat "" (Array.to_list (Array.map (fun (id', prm') -> nm ^ (*".headnames[" ^ nm ^ ".headnames.length] = \"" ^*)
																										".headnames." ^ id' ^ " = " ^ nm ^ 
																										".headtypes.length;\n" ^ nm ^ ".headtypes[" ^ nm ^ ".headtypes.length] = \"" ^
																										(match prm' with Int -> "int" | Float -> "float" | String -> "string") ^
																										"\";\n") 
																									(match tb_sig with Signature(sigsig) -> sigsig)
			)))
			^ nm ^ ".rows[" ^ nm ^ ".rows.length] = new Array();\n" ^
			(String.concat (nm ^ ".rows[" ^ nm ^ ".rows.length] = new Array();\n") 
				(Array.to_list (Array.map 
					(fun mlitrow -> let mlitarray = match mlitrow with LiteralRow(lits) -> lits in 
						String.concat "" (Array.to_list (Array.map 
																(fun mlit -> nm ^ ".rows[" ^ nm ^ ".rows.length - 1]" ^
																"[" ^ nm ^ ".rows[" ^ nm ^ ".rows.length - 1].length] = " ^ 
																(match mlit with LitInt(i) -> string_of_int i | LitFloat(f) -> string_of_float f | LitString(s) -> s)
					 										^ ";\n")
					 	 										mlitarray 
					 	 									))
					) 
					tablerows
			)))			 
		in javaout :: translate_globals tl

	in	


	let translate thisenv fundecl =
			(* Now we have our typecheck function *)
		let () = thisenv.current_enclosing_function <- fundecl.fname
			(*; print_string ("current_enclosing_function set: " ^ fundecl.fname)*) in
		let rec typecheck_of_expr = function
    	Literal(l) -> (match l with LitInt(i) -> SingleVarLiteral(("",Int),LitInt(0)) 
    													| LitFloat(f) -> SingleVarLiteral(("",Float),LitFloat(0.0)) 
    													| LitString(s) -> SingleVarLiteral(("",String),LitString("")))
  		| Id(s) -> (try (match (StringMap.find s thisenv.global_index) with 
  											SingleVarLiteral((_,Int),_) -> SingleVarLiteral(("",Int),LitInt(0))
  										| SingleVarLiteral((_,Float),_) -> SingleVarLiteral(("",Float), LitFloat(0.0))
  										| SingleVarLiteral((_,String),_) -> SingleVarLiteral(("",String), LitString(""))
  										| FullTableLiteral(_, sig_n, _) -> FullTableLiteral("", sig_n, [||])
  										) 
  							 
  								with Not_found -> (try (match (StringMap.find s thisenv.local_index) with 
  											(SingleVarLiteral((_,Int),_),_) -> SingleVarLiteral(("",Int),LitInt(0))
  										| (SingleVarLiteral((_,Float),_),_) -> SingleVarLiteral(("",Float), LitFloat(0.0))
  										| (SingleVarLiteral((_,String),_),_) -> SingleVarLiteral(("",String), LitString(""))
  										| (FullTableLiteral(_, sig_n, _),_) -> FullTableLiteral("", sig_n, [||])
  										) 
								 										 with Not_found -> raise(Failure("Unknown id: " ^ s))
																		)
									)
		  | Binop(e1, o, e2) -> let p1 = (match (typecheck_of_expr e1) with SingleVarLiteral((_,stprm),_) -> Signature([|("$sv",stprm)|])
		  																																|FullTableLiteral(_,sig_1,_) -> sig_1 
		  																)
		  											and p2 = (match (typecheck_of_expr e2) with SingleVarLiteral((_,stprm2),_) -> Signature([|("$sv",stprm2)|])
		  																																|FullTableLiteral(_,sig_2,_) -> sig_2 
		  																)
		  											in if(p1 <> p2) then (
  			raise(Failure("\nType mismatch of expression: \n" ^ (Ast.string_of_expr e1) ^ "\n***\n" ^ (Ast.string_of_expr e2)))) else (
      		typecheck_of_expr e2
				)
  		| Assign(v, e) -> let p1 = (try (match (StringMap.find v thisenv.global_index) with 
  																						SingleVarLiteral((_,stprm),_) -> Signature([|("sv",stprm)|])
		  																				| FullTableLiteral(_,sig_1,_) -> sig_1 
		  																) with Not_found -> (try (match(StringMap.find v thisenv.local_index) with
		  																				(SingleVarLiteral((_,stprm),_),_) -> Signature([|("sv",stprm)|])
		  																				| (FullTableLiteral(_,sig_1,_),_) -> sig_1 
		  																				)
								 										 with Not_found -> raise(Failure("Unknown id: " ^ v))))
		  										and p2 = (match (typecheck_of_expr e) with SingleVarLiteral((_,stprm2),_) -> Signature([|("sv",stprm2)|])
		  																																|FullTableLiteral(_,sig_2,_) -> sig_2 
		  															)
		  										in if(p1 <> p2) then (
  			raise(Failure("\nType mismatch in assignment: \n" ^ v ^ "\n***\n" ^ (Ast.string_of_expr e)))) else (
      		typecheck_of_expr e
				)
			| TAssign(v, v', e) -> let p1 = typecheck_of_expr (GetCol(v, v'))
														 and p2 = typecheck_of_expr e
														 in if p1 <> p2 then raise(Failure("types do not match in TAssign")) else typecheck_of_expr e
		  | Call(f, el) -> (try(if(List.for_all2 (fun (_,p) (_,p') -> p = p') 
		  																			(let f_formals = StringMap.find f thisenv.function_index in f_formals.formals) 
		  																			(List.map (fun tabledec -> match tabledec with SingleVarLiteral(sp,_) -> sp
		  																				| s -> raise(Illegal_table_param(s))) 
		  																				(List.map typecheck_of_expr el)
		  																			)
		  											 )then(
		  											 	let func = (StringMap.find f thisenv.function_index) in func.returnsig
		  											 )else(raise(Failure("sequence of types supplied to " ^ f ^ " does not equal expected types"))
		  										)) with Not_found -> raise(Function_not_found(f))
		  											| Illegal_table_param(s) -> (if(StringMap.mem f thisenv.function_index)
		  												then(raise(Failure("\n Can't pass a table as parameter: " ^ Ast.string_of_tdecl s ^ "\n")))
		  												else(raise(Function_not_found(f)))
		  																									)
		  									)
		  | Noexpr -> raise(Failure("A NoExpr has no type; it cannot be type checked."))
			| GetCol(s1, s2) -> let p1 = (try (match (StringMap.find s1 thisenv.global_index) with 
  												SingleVarLiteral((_,stprm),_) -> raise(Failure("Can't get a column of a global primitive.\n"))
		  									| FullTableLiteral(_,Signature(sig_1),sigarry) -> let () = if(Array.length sigarry <> 1) then(
		  											raise(Failure(s1 ^ ", a global, has " ^ string_of_int (Array.length sigarry) ^ " rows, can't refer to a single element"))) else (
		  											print_string "WARNING: trying to access a column of a global table\n") 
		  										in sig_1 
		  															) with Not_found -> (try (match(StringMap.find s1 thisenv.local_index) with
		  										(SingleVarLiteral((_,stprm),_),_) -> raise(Failure("Can't get a column of a local primitive.\n"))
		  									| (FullTableLiteral(_,Signature(sig_1),sigarry),_) -> let () = if(Array.length sigarry <> 1)then(
		  											raise(Failure(s1 ^ ", a local, has " ^ string_of_int (Array.length sigarry) ^ " rows, can't refer to a single element")))else() in sig_1
		  																												)
								 											with Not_found -> raise(Failure("Unknown id: " ^ s1))
								 																				)
								 								)
		  								in if(List.exists (fun elt -> fst elt = s2) (Array.to_list p1)) then (
		  									(* get the type associated with the reference*)
		  									(match (List.find (fun elt -> fst elt = s2) (Array.to_list p1)
		  												 ) with (_,Int) -> SingleVarLiteral(("",Int),LitInt(0))
		  															| (_,Float) -> SingleVarLiteral(("",Float),LitFloat(0.0))
		  															| (_,String) -> SingleVarLiteral(("",String),LitString(""))
		  									)
		  								) else(
  											raise(Failure("'" ^ s2 ^ "' is not a valid signature element."))
  										)
(*  | Filter(s1, ex2) -> s1 ^ "[" ^ javaS_of_expr ex2 ^ "]"
*)

in
	let rec javaS_of_expr = function
    Literal(l) -> (match l with LitInt(i) -> string_of_int i | LitFloat(f) -> string_of_float f | LitString(s) -> s)
  | Id(s) -> s
  | Binop(e1, o, e2) as binexp -> (* Here we need to typecheck each expression before generating the code for it.*)
      (try (match (typecheck_of_expr binexp) with 
      SingleVarLiteral(( _, Int), _) -> javaS_of_expr e1 ^
     		(match o with Add -> " + " | Sub -> " - " | Mult -> " * " | Div -> " / " | Equal -> " == " | Neq -> " != "
      	| Less -> " < " | Leq -> " <= " | Greater -> " > " | Geq -> " >= " (* | And_ -> "&&" | Or_ -> "||"*)
      	) ^ javaS_of_expr e2
      | SingleVarLiteral((_, Float),_) -> javaS_of_expr e1 ^
     		(match o with Add -> " + " | Sub -> " - " | Mult -> " * " | Div -> " / " | Equal -> " == " | Neq -> " != "
      	| Less -> " < " | Leq -> " <= " | Greater -> " > " | Geq -> " >= " (* | And_ -> "&&" | Or_ -> "||"*)
      	) ^ javaS_of_expr e2
      | SingleVarLiteral((_, String),_) -> javaS_of_expr e1 ^
     		(match o with Add -> " + " | Equal -> " == " | Neq -> " != " | Less -> " < " | Leq -> " <= " | Greater -> " > "
     		| Geq -> ">= " | _ -> raise(Failure("Can't perform this operation on a string"))(* | And_ -> "&&" | Or_ -> "||"*)
      	) ^ javaS_of_expr e2
      | FullTableLiteral(_,_,_) -> javaS_of_expr e1 ^ ".signature" ^ 
      	(match o with Equal -> " == " | Neq -> " != " | _ -> raise(Failure("Can't perform this operation on a table"))
      	(* | And_ -> "&&" | Or_ -> "||"*)
      	) ^ javaS_of_expr e2 ^ ".signature"

					) with Failure(msg) -> raise(Failure("Can't generate code: " ^ msg ^ "\n"))
			)
  | Assign(v, e) as assignexp -> (try (match(typecheck_of_expr assignexp) with
  		_ -> v ^ " = " ^ javaS_of_expr e) with Failure(msg) -> raise(Failure("Can't generate code: " ^ msg ^ "")))
  | TAssign(v, v', e) as assignexp -> (try (match(typecheck_of_expr assignexp) with
  		_ -> javaS_of_expr (GetCol(v, v')) ^ " = " ^ javaS_of_expr e) with Failure(msg) -> raise(Failure("Can't generate code: " ^ msg ^ "")))
  | Call(f, el) as callexp -> (try (match(typecheck_of_expr callexp) with
  		_ -> f ^ "(" ^ (String.concat ", " (List.map javaS_of_expr el)) ^ ")") 
  			with Invalid_argument(msg) -> raise(Failure("Call to function " ^ f ^ " does not match the declared argument list for that function: " ^ msg)) 
		  	| Function_not_found(func) -> (* Code for standard library.*) 
		  	(match f with 
		  		"print" -> "alert(" ^ (if(List.length el = 1)then(match typecheck_of_expr(List.hd el) with
		  				SingleVarLiteral(_,_) -> javaS_of_expr (List.hd el)
		  			| FullTableLiteral(_,_,_) -> raise(Failure("Call toDOM on a table instead of print."))
		  			)else(raise(Failure("Wrong # of parameters passed to print")))) ^ ")"
		  	| "toDOM" -> "$$tml$$.toDOM(" ^ (if(List.length el = 1)then(match typecheck_of_expr(List.hd el) with
		  				SingleVarLiteral(_,_) -> raise(Failure("Call print on a primitive instead of toDOM."))
		  			| FullTableLiteral(_,_,_) -> javaS_of_expr (List.hd el)
		  			)else(raise(Failure("Wrong # of parameters passed to toDOM")))) ^ ")"
		  	| "delete" -> (if(List.length el = 1)then(match typecheck_of_expr(List.hd el) with
		  				SingleVarLiteral(_,_) -> let () = print_string "WARNING: deleting a primitive will have no effect\n" 
		  					in "delete " ^ javaS_of_expr (List.hd el)
		  			| FullTableLiteral(_,_,_) -> "$$tml$$.del(" ^ javaS_of_expr (List.hd el)
		  			)else(raise(Failure("Wrong # of parameters passed to toDOM")))) ^ ")"
		  	| _ -> raise(Failure("Function " ^ func ^ " not found"))
		  	)
  			)(*End try *)
  | Noexpr -> ""
  (* first thing in GetCol is to call the typechecker.  After running that function, you can be confident that s1 exists, that s2
  	refers to a valid element.*)
  | GetCol(s1, s2) as getcolexp -> try(match(typecheck_of_expr getcolexp) with _ -> s1 ^ ".rows[0][" ^ s1 ^ ".headnames." ^ s2 ^ "]" ) 
  																 with Failure(msg) -> raise(Failure("Can't generate code: " ^ msg))
(*  | Filter(s1, ex2) -> s1 ^ "[" ^ javaS_of_expr ex2 ^ "]" *)

	in let rec javaS_of_stmt = function
    Block(stmts) -> (*let () = print_string "HEYME" in *)
      "{\n" ^ String.concat "" (List.map javaS_of_stmt stmts) ^ "}\n"
  | Expr(expr) -> javaS_of_expr expr ^ ";\n"
  | Return(expr) -> (* If the typecheck of the expression doesn't match the returnsig of the enclosing function, then throw an error*)
  		let () = 
  			let encl_func = (StringMap.find thisenv.current_enclosing_function thisenv.function_index) in 
  			if encl_func.returnsig <> typecheck_of_expr expr then 
  				raise(Failure("Return statement : " ^ Ast.string_of_expr expr ^ " does not match return statement of: " ^ 
  					thisenv.current_enclosing_function)) 
  			else ()
  		in "return " ^ javaS_of_expr expr ^ ";\n"
  | If(e, s, Block([])) -> "if (" ^ javaS_of_expr e ^ ")\n" ^ javaS_of_stmt s
  | If(e, s1, s2) ->  "if (" ^ javaS_of_expr e ^ ")\n" ^
      javaS_of_stmt s1 ^ "else\n" ^ javaS_of_stmt s2
  | For(e1, e2, e3, s) ->
      "for (" ^ javaS_of_expr e1  ^ " ; " ^ javaS_of_expr e2 ^ " ; " ^
      javaS_of_expr e3  ^ ") " ^ javaS_of_stmt s
  | While(e, s) -> "while (" ^ javaS_of_expr e ^ ") " ^ javaS_of_stmt s
  | Pipe(s1, s2, stmts) -> let () = if (try(match (StringMap.find s1 thisenv.global_index) with
  																						FullTableLiteral(_,_,arry) -> Array.length arry
  																					| SingleVarLiteral(_,_) -> raise(Failure(s1 ^ " is a global primitive, not a table\n"))
  																				 ) with Not_found -> try (match (StringMap.find s1 thisenv.local_index) with
  																						(FullTableLiteral(_,_,arry),_) -> Array.length arry
  																					| (SingleVarLiteral(_,_),_) -> raise(Failure(s1 ^ " is a local primitive, not a table\n"))
  																				 ) with Not_found -> raise(Failure( s1 ^ " is not a valid table variable.\n"))
  																			) = 0 then raise(Failure(s1 ^ " does not have any rows to iterate through.\n")) 
  																	else (try(match (StringMap.find s1 thisenv.global_index) with
  																					(*	FullTableLiteral(_,_,_) as ftl -> thisenv.local_index <- StringMap.add s2 (ftl,(StringMap.find thisenv.current_enclosing_function thisenv.function_index)) thisenv.local_index *)
  																						FullTableLiteral(name, sig_n, litrowarry) -> let newarry = [|litrowarry.(0)|] in
  																							thisenv.local_index <- StringMap.add s2 (FullTableLiteral(name, sig_n, newarry), (StringMap.find thisenv.current_enclosing_function thisenv.function_index)) thisenv.local_index
  																					| SingleVarLiteral(_,_) -> raise(Failure(s1 ^ " is a global primitive, not a table\n"))
  																				 ) with Not_found -> try (match (StringMap.find s1 thisenv.local_index) with
  																						(*(FullTableLiteral(_,_,_),_) as ftl -> thisenv.local_index <- StringMap.add s2 ftl thisenv.local_index *)
  																						(FullTableLiteral(name, sig_n, litrowarry),encl_func) -> let newarry = [|litrowarry.(0)|] in thisenv.local_index <- StringMap.add s2 (FullTableLiteral(name, sig_n, newarry),encl_func) thisenv.local_index 
  																					| (SingleVarLiteral(_,_),_) -> raise(Failure(s1 ^ " is a local primitive, not a table\n"))
  																				 ) with Not_found -> raise(Failure( s1 ^ " is not a valid table variable.\n"))
  																			 )
  	(*let () = print_string (s2 ^ " points to table " ^ s1 ^ "\n") in *) 
  		in "var " ^ s2 ^ " = {};\nfor (var $" ^ s1 ^ " = 0; $" ^ s1 ^ " < " ^ s1 ^ ".rows.length; $" ^ s1 ^ "++ ) {\n" ^ 
  			s2 ^ ".rows = new Array();\n" ^
  			s2 ^ ".rows[0] = " ^ s1 ^ ".rows[$" ^ s1 ^ "];\n" ^ s2 ^ ".signature = " ^ s1 ^ ".signature;\n" ^
  			s2 ^ ".headnames = " ^ s1 ^ ".headnames;\n" ^ s2 ^ ".headtypes = " ^ s1 ^ ".headtypes;\n" ^
  			(String.concat "" (List.map javaS_of_stmt stmts)) ^ "}\n"
  | LocalDecl(td) -> (match td with SingleVarLiteral((s, _), _) as svl ->
  	let () = (*print_string ("found variable: " ^ s ^ "::" ^ thisenv.current_enclosing_function ^ "\n");*) thisenv.local_index <- StringMap.add s (svl,(StringMap.find thisenv.current_enclosing_function thisenv.function_index)) thisenv.local_index
  	in String.concat "" (translate_globals [td])
  										| FullTableLiteral(s,_,_) as ftl ->
  				let () = thisenv.local_index <- StringMap.add s (ftl,
  				(StringMap.find thisenv.current_enclosing_function thisenv.function_index)) thisenv.local_index
  	in String.concat "" (translate_globals [td])
  									 )
  | ColAppend(s1, s2) -> let p1 = (try (match (StringMap.find s1 thisenv.global_index) with 
													FullTableLiteral(_,_,_) as ftlit -> (ftlit, "GL")
  											|	_ -> raise(Failure("Can't append a column to a global primitive.\n"))
		  															) with Not_found -> (try (match(StringMap.find s1 thisenv.local_index) with
		  										(FullTableLiteral(_,_,_) as ftlit,encl_func) -> (ftlit, encl_func.fname)
		  									|	_ -> raise(Failure("Can't append a column to a local primitive.\n"))
		  																												)
								 											with Not_found -> raise(Failure("Unknown ColAppend id: " ^ s1))
								 																				)
								 								)
								 				and p2 = (try (StringMap.find s2 thisenv.global_index) with 
  																	Not_found -> (try (match(StringMap.find s2 thisenv.local_index) with
		  										(ftlit2,_) -> ftlit2
		  																								)
								 								with Not_found -> raise(Failure("Unknown ColAppend id: " ^ s2))
								 																 )
								 								)
		  			in let p1parts = (match (fst p1) with FullTableLiteral(_,Signature(sig_n),litrows) -> (sig_n, litrows)
		  																						| _ -> raise(Failure("NEVER CALLED")))
		  			and p2parts = (match p2 with FullTableLiteral(_,Signature(sig_n2),litrows2) -> (sig_n2, litrows2.(0))
		  																	| SingleVarLiteral(sig_n2,onelit) -> ([|sig_n2|], LiteralRow([|onelit|])))
		  			in let newsigarray = Array.append (fst p1parts) (fst p2parts)
		  			and litstoappend = Array.to_list (match (snd p2parts) with LiteralRow(litrowarray) -> litrowarray)
		  			in let newtablelit = Array.of_list(List.map 
		  				(fun litrow -> LiteralRow(Array.of_list((Array.to_list(match litrow with LiteralRow(litrowarray2) -> litrowarray2)) @ litstoappend))) 
		  					(Array.to_list (snd p1parts)))
		  			in let () = if (snd p1) = "GL" then thisenv.global_index <- (StringMap.add s1 (FullTableLiteral(s1, Signature(newsigarray), newtablelit)) thisenv.global_index) else thisenv.local_index <- (StringMap.add s1 ((FullTableLiteral(s1, Signature(newsigarray), newtablelit)),(StringMap.find (snd p1) thisenv.function_index)) thisenv.local_index)
  				
					in 
					"if(" ^ s2 ^ ".rows) {\n" ^ 
					"for(var $" ^ s2 ^ " in " ^ s2 ^ ".headnames) {\n" ^
					"if(" ^ s1 ^ ".headnames[$" ^ s2 ^ "]) {\nthrow \"Duplicate Column header.\";\n}\nelse {\n" ^
					s1 ^ ".headnames[$" ^ s2 ^ "] = " ^ s1 ^ ".headtypes.length;\n" ^
					s1 ^ ".headtypes.push(" ^ s2 ^ ".headtypes[" ^ s2 ^ ".headnames[$" ^ s2 ^ "]]);\n}\n" ^
					"}\n" ^
					s1 ^ ".signature += \"::\" + " ^ s2 ^ ".signature;\n" ^
      		"}\n else {" ^
      		s1 ^ ".headnames." ^ s2 ^ " = " ^ s1 ^ ".headtypes.length;\n" ^
					s1 ^ ".headtypes.push(\"" ^ 
					(match (fst p2parts).(0) with (_,prm) -> (match prm with Int -> "int" | Float -> "float" | String -> "string")) ^ 
					"\");\n" ^ s1 ^ ".signature += \"::" ^ s2 ^ ":" ^ 
					(match (fst p2parts).(0) with (_,prm) -> (match prm with Int -> "int" | Float -> "float" | String -> "string")) ^ 
					"\";\n" ^"}\n" ^
      		"for (var $" ^ s1 ^ " = 0; $" ^ s1 ^ " < " ^ s1 ^ ".rows.length; $" ^ s1 ^ "++ ) {\n" ^ 
      		"if(" ^ s2 ^ ".rows) {\n" ^ s1 ^ ".rows[$" ^ s1 ^ "] = " ^ s1 ^ ".rows[$" ^ s1 ^ "].concat(" ^ s2 ^ ".rows[0]);\n" ^
      		"}\n else {" ^ s1 ^ ".rows[$" ^ s1 ^ "].push(" ^ s2 ^ ");\n}" ^ "\n}\n"

	| RowAppend(s1,s2) -> let p1 = (try (match (StringMap.find s1 thisenv.global_index) with 
													FullTableLiteral(_,Signature(sig_1),litarry) -> (sig_1, litarry)
  											|	_ -> raise(Failure("Can't append a row to a global primitive.\n"))
		  															) with Not_found -> (try (match(StringMap.find s1 thisenv.local_index) with
		  										(FullTableLiteral(_,Signature(sig_1),litarry),_)  -> (sig_1, litarry)
		  									|	_ -> raise(Failure("Can't append a row to a local primitive.\n"))
		  																												)
								 											with Not_found -> raise(Failure("Unknown rowAppend id: " ^ s1))
								 																				)
								 								)
								 				and p2 = (try (match (StringMap.find s2 thisenv.global_index) with 
  												FullTableLiteral(_,Signature(sig_2),litarry2) -> (sig_2,litarry2)
  											| _ -> raise(Failure("Can't append a row to a global primitive.\n"))
		  															) with Not_found -> (try (match(StringMap.find s2 thisenv.local_index) with
		  										(FullTableLiteral(_,Signature(sig_2),litarry2),_) -> (sig_2,litarry2)
		  									| _ -> raise(Failure("Can't append a row to a local primitive.\n"))
		  																												)
								 											with Not_found -> raise(Failure("Unknown rowAppend id: " ^ s2))
								 																				)
								 								)
		  								in if(fst p1 <> fst p2) then ( 
  			raise(Failure("\nType mismatch of rowAppend: \n" ^ s1 ^ "\n***\n" ^ s2))) else (
  				let () = (try (match (StringMap.find s1 thisenv.global_index) with 
													FullTableLiteral(s,sig_n,_) -> (thisenv.global_index <- 
															(StringMap.add s1 (FullTableLiteral(s, sig_n, (Array.append (snd p1) (snd p2)))) thisenv.global_index))
  											|	_ -> ()
		  															) with Not_found -> (try (match(StringMap.find s1 thisenv.local_index) with
		  										(FullTableLiteral(s,sig_n,_),encl_func) -> (thisenv.local_index <- 
															(StringMap.add s1 (FullTableLiteral(s, sig_n, (Array.append (snd p1) (snd p2))), encl_func) thisenv.local_index))
		  									|	_ -> ()
		  																												)
								 											with _ -> ()
								 																				)
								 								) in
      		"for (var $" ^ s2 ^ " = 0; $" ^ s2 ^ " < " ^ s2 ^ ".rows.length; $" ^ s2 ^ "++ ) {\n" ^ 
  				s1 ^ ".rows.push($$tml$$.cpy(" ^ s2 ^ ".rows[$" ^ s2 ^ "]));\n}\n"
				)
	in
		let () = for i = 0 to (List.length fundecl.formals) - 1 do 
			let localvars = Array.of_list fundecl.formals 
			in match localvars.(i) with 
				(s, Int) -> thisenv.local_index <- (StringMap.add s (SingleVarLiteral((s, Int), LitInt(0)),fundecl) thisenv.local_index)
			|	(s, Float) -> thisenv.local_index <- (StringMap.add s (SingleVarLiteral((s, Float), LitFloat(0.0)),fundecl) thisenv.local_index)
			| (s, String) -> thisenv.local_index <- (StringMap.add s (SingleVarLiteral((s, String), LitString("")),fundecl) thisenv.local_index)
		done
		in
		let func_string = "function " ^ fundecl.fname ^ "(" ^ (*(let () = print_string thisenv.current_enclosing_function in "") ^*)  
		(String.concat ", " (List.map (fun (s, p) -> s) fundecl.formals)) ^ ")\n{\n" ^ (*(let () = print_string "POP" in "") ^ *)
		(String.concat "" (List.map javaS_of_stmt fundecl.body)) ^ "\n}\n" (*^ (let () = print_string "GOES" in "") *) 
		(* Now remove all the local variables that were declared in this function *)
		in
		let () = (*print_string "BBB";*) StringMap.iter 
					(fun key (_, encl_func) -> if(thisenv.current_enclosing_function = encl_func.fname)then(
						thisenv.local_index <- StringMap.remove key thisenv.local_index
					)else()
					) thisenv.local_index(*; print_string "AAA"*)
		in func_string
		(* End of declarations within translate, now time to set environment and call the function *)
		in let thisenv = { function_index = global_func_decls;
											 global_index = global_table_decls;
											 local_index = StringMap.empty;
											 current_enclosing_function = "";
								}
	in (* First we need to import the standard library *)
		("if(!$$tml$$) { throw new Error(\"STANDARD LIBRARY NOT FOUND\"); }\n") ^ 
		(* Next we translate globals, then functions *)
		(String.concat "" (List.rev (translate_globals globals))) ^ String.concat "" (List.map (translate thisenv) functions)
		(* If it's the main function, invoke it *)
		^ (if List.exists (fun fnc ->if(fnc.fname = "main")then true else false) functions then "\nmain();\n" else 
				raise(Failure("No main function found")))

